;
;   Module name      :	IniUtil.au3
;   Version          :	1.1
;   Started on       :	June 9, 2013
;   Completed on     :	June 9, 2013
;   Author           :  J.Hubb
;
;   Module type      :	Library
;   Description      :	Utilities for handling .ini files
;
;   Developed under  :	AutoIt 3.2.10.0
;   Sw requirements  :	None
;
;   Parameters       :	None
;   Notes - usage    :	The library exports 5 functions for .ini handling:
;							IUInit			Searches and returns the complete path of a requested .ini
;											inside a list of paths
;							IUReadArrays	Reads groups of variable values from .ini sections and assigns
;											them to the specified global arrays
;							IUReadVars		Reads variable values from .ini sections and assigns them to
;											the specified global variables
;							IUWriteArrays	Writes global arrays to specified groups of variables in sections
;											of .ini
;							IUWriteVars		Writes global variables to specified variables in sections of .ini
;


;--- Global directives -----------------------------------------------------------------------------



#include-once

Opt("MustDeclareVars", 1)



;--- Include files ---------------------------------------------------------------------------------



;--- Public constants ------------------------------------------------------------------------------



const $IU_FILE_NAME = "IniUtil.au3"
const $IU_MODULE_VER = "1.1"
const $IU_MODULE_DATE = "June 9, 2013"
const $IU_MODULE_NAME = $IU_FILE_NAME & " " & $IU_MODULE_VER & "; Open Source library on " & $IU_MODULE_DATE & " by J.Hubb."

; Error codes
const $IU_NO_ERROR = 0
const $IU_NOT_FOUND = 1
const $IU_INVALID_DEF = 2
const $IU_UNDEF_VAR = 3
const $IU_VAR_NOT_FOUND = 4
const $IU_WRITE_FAILED = 5



;--- Public variables ------------------------------------------------------------------------------



;--- Private constants -----------------------------------------------------------------------------



;--- Private variables -----------------------------------------------------------------------------



;--- Public functions code -------------------------------------------------------------------------



func IUInit($name = "", $search = "")
;---
; Searches the specified name $name inside $search list of paths; if not found, @error is set.
; If $name includes a path, then no search but just an existence check on $name is performed.
; If $name includes an extension, this is substituted with ".ini".
;
; $name		ini file name with or without path; if not specified, the script name is used
; $search	list of search path for .ini; elements are separated by ";"; if not specified, the
;			default search path is: current dir., script dir., System Root dir. (= WINDOWS)
; ret		complete pathname found for $name
;---
	local $pos, $s_list, $d
	
	if ($name == "") then
		$name = @ScriptName
	endif
	$pos = StringInStr($name, ".", 2, -1)
	if ($pos > 0) then
		$name = StringLeft($name, $pos - 1) & ".ini"
	endif
	if (StringInStr($name, "\", 2) > 0) then
		if (FileExists($name)) then
			return $name
		else
			SetError($IU_NOT_FOUND)
			return $name
		endif
	else
		if ($search == "") then
			$search = ".;" & @ScriptDir & ";" & @WindowsDir
		endif
		$s_list = StringSplit($search, ";")
		for $d in $s_list
			if (FileExists(_IUAddTrailingBS($d) & $name)) then
				return _IUAddTrailingBS($d) & $name
			endif
		next
		SetError($IU_NOT_FOUND)
		return $name
	endif
endfunc



func IUReadArrays(const $ini, const byref $def)
;---
; Reads from $ini file, values from .ini to global arrays as specified in $def; on errors, @error is set
; and details are returned in an array.
;
; $ini	ini file path
; $def	array with definitions of .ini variables and global arrays; every entry is an inner array in the form:
;			[IniSectionName, IniVarNameStruc, GlobalArrayName, DefaultValue]
;		where:
;			IniSectionName	name of the .ini section to be used
;			IniVarNameStruc	structure of names of the .ini variables inside IniSectionName (details below)
;			GlobalArrayName	name of the global array to be assigned (must already exists)
;			DefaultValue	default value to be assigned to GlobalVarName if IniVarName is not found
;		If $def has uncorrect dimensions or name structure, the function sets @error to $IU_INVALID_DEF and
;		returns.
;		If GlobalArrayName has not already been defined, the function sets @error to $IU_UNDEF_VAR and returns.
;		If DefaultValue is "default" and one variable name defined in IniVarNameStruc is not found, the
;		function sets @error to $IU_VAR_NOT_FOUND and returns.
;		The entry IniVarNameStruc is list of elements separater by "=" ("=" is a senseless char in .ini
;		variable name); it can have 2 forms:
;			"List=VarName1=VarName2=VarName3=..." where VarName<n> are variable names that should be read
;					in appearing order and assigned to array elements 0..<n-1>; the list should end with
;					a VarName, not with a "="
;			"Group=VarName" where VarName is the variable name for all rows with the same name to be read;
;					count of lines is determined from .ini section
;			"Step=VarNamePrefix=DigitsNum=StartNum=EndNum" where VarNamePrefix is the first part of variable
;					name at which is appended a number composed by DigitsNum digits, starting at StartNum
;					value and ending at EndNum value included
;		On errors on IniVarNameStruc structure, the function sets @error to $IU_INVALID_DEF and returns.
; ret	if no error occurred			1
;		if @error = $IU_INVALID_DEF		["InnerDim", <n>] where <n> is used dims of inner array (should be 4)
;										["TooShort", <str>] where <str> is the invalid structure used
;										["InvStruct", <str>] where <str> is the invalid structure used
;										["InvOper", <oper>] where <oper> is the invalid operation (not "List",
;															nor "Step")
;		if @error = $IU_UNDEF_VAR		[<name>, ""] where <name> is the global variable name not defined
;		if @error = $IU_VAR_NOT_FOUND	[<sect>, <name>] where <sect> is the .ini section where the variable
;														<name> has not been found
;---
	local $tmp, $one_def, $elem_count, $oper, $n, $i, $tmp_arr[1], $ini_sect_arr[1][1]
	local $prefix, $digits, $start, $end, $ini_var_name
	local enum $_IU_SECT, $_IU_INI_ARR, $_IU_GLOB_ARR, $_IU_DEF		; Positions in inner arrays

	$tmp = UBound($def, 2)
	if ($tmp <> 4) then
		local $res[2] = ["InnerDim", $tmp]
		SetError($IU_INVALID_DEF)
		return $res
	endif
	for $n = 0 to (UBound($def) - 1)
		$one_def = StringSplit($def[$n][$_IU_INI_ARR], "=")
		$elem_count = $one_def[0]
		if ($elem_count == 1) then
			local $res[2] = ["TooShort", $def[$n][1]]
			SetError($IU_INVALID_DEF)
			return $res
		endif
		if (IsDeclared($def[$n][$_IU_GLOB_ARR]) <> 1) then
			local $res[2] = [$def[$n][$_IU_GLOB_ARR], ""]
			SetError($IU_UNDEF_VAR)
			return $res
		endif
		$oper = StringLower($one_def[1])
		switch $oper
			case "list"
				if ($elem_count < 2) then
					local $res[2] = ["TooShort", $def[$n][1]]
					SetError($IU_INVALID_DEF)
					return $res
				endif
				ReDim $tmp_arr[$elem_count - 1]			; The first is "List"
				for $i = 2 to $elem_count
					$ini_var_name = $one_def[$i]
					$tmp = IniRead($ini, $def[$n][$_IU_SECT], $ini_var_name, $def[$n][$_IU_DEF])
					if ($tmp == default) then
						local $res[2] = [$def[$n][$_IU_SECT], $ini_var_name]
						SetError($IU_VAR_NOT_FOUND)
						return $res
					endif
					$tmp_arr[$i - 2] = $tmp
				next
				Assign($def[$n][$_IU_GLOB_ARR], $tmp_arr, 6)	; 2 + 4 = global scope + fail if not existant
				Dim $tmp_arr[1]									; Make $tmp_arr empty to save memory
			case "group"
				if ($elem_count < 2) then
					local $res[2] = ["TooShort", $def[$n][1]]
					SetError($IU_INVALID_DEF)
					return $res
				endif
				$ini_sect_arr = IniReadSection($ini, $def[$n][$_IU_SECT])
				if (@error == 1) then							; The requested section does not exist
					if ($def[$n][$_IU_DEF] == default) then
						local $res[2] = [$def[$n][$_IU_SECT], "*"]
						SetError($IU_VAR_NOT_FOUND)
						return $res
					else
						$tmp_arr[0] = $def[$n][$_IU_DEF]
						Assign($def[$n][$_IU_GLOB_ARR], $tmp_arr, 6)	; 2 + 4 = global scope + fail if
						return 1										; not existant
					endif
				endif
				$elem_count = $ini_sect_arr[0][0]
				ReDim $tmp_arr[$elem_count]
				$ini_var_name = $one_def[2]
				$start = 0
				for $i = 1 to $elem_count
					if ($ini_sect_arr[$i][0] == $ini_var_name) then
						$tmp_arr[$start] = $ini_sect_arr[$i][1]
						$start += 1
					endif
				next
				if ($start == 0) then
					if ($def[$n][$_IU_DEF] == default) then
						local $res[2] = [$def[$n][$_IU_SECT], $ini_var_name]
						SetError($IU_VAR_NOT_FOUND)
						return $res
					else
						$tmp_arr[0] = $def[$n][$_IU_DEF]
						$start = 1
					endif
				endif
				ReDim $tmp_arr[$start]
				Assign($def[$n][$_IU_GLOB_ARR], $tmp_arr, 6)	; 2 + 4 = global scope + fail if not existant
				Dim $tmp_arr[1], $ini_sect_arr[1][1]			; Make arrays empty to save memory
			case "step"
				if ($elem_count <> 5) then
					local $res[2] = ["InvStruct", $def[$n][1]]
					SetError($IU_INVALID_DEF)
					return $res
				endif
				$prefix = $one_def[2]
				$digits = $one_def[3]
				$start = $one_def[4]
				$end = $one_def[5]
				ReDim $tmp_arr[$end - $start + 1]
				for $i = $start to $end
					$ini_var_name = StringFormat("%s%0" & $digits & "u", $prefix, $i)
					$tmp = IniRead($ini, $def[$n][$_IU_SECT], $ini_var_name, $def[$n][$_IU_DEF])
					if ($tmp == default) then
						local $res[2] = [$def[$n][$_IU_SECT], $ini_var_name]
						SetError($IU_VAR_NOT_FOUND)
						return $res
					endif
					$tmp_arr[$i - $start] = $tmp
				next
				Assign($def[$n][$_IU_GLOB_ARR], $tmp_arr, 6)	; 2 + 4 = global scope + fail if not existant
				Dim $tmp_arr[1]									; Make $tmp_arr empty to save memory
			case else
				local $res[2] = ["InvOper", $oper]
				SetError($IU_INVALID_DEF)
				return $res
		endswitch
	next
	return 1
endfunc



func IUReadVars(const $ini, const byref $def)
;---
; Reads from $ini file, values from .ini to global variables as specified in $def; on errors, @error is set
; and details are returned in an array.
;
; $ini	ini file path
; $def	array with definitions of .ini and global variables; every entry is an inner array in the form:
;			[IniSectionName, IniVarName, GlobalVarName, DefaultValue]
;		where:
;			IniSectionName	name of the .ini section to be used
;			IniVarName		name of the .ini variable inside IniSectionName
;			GlobalVarName	name of the global variable to be assigned (must already exists)
;			DefaultValue	default value to be assigned to GlobalVarName if IniVarName is not found
;		If $def has uncorrect dimensions, the function sets @error to $IU_INVALID_DEF and returns.
;		If GlobalVarName has not already been defined, the function sets @error to $IU_UNDEF_VAR and returns.
;		If DefaultValue is "default" and IniVarName is not found, the function sets @error to $IU_VAR_NOT_FOUND
;		and returns.
; ret	if no error occurred			1
;		if @error = $IU_INVALID_DEF		["InnerDim", <n>] where <n> is used dims of inner array (should be 4)
;		if @error = $IU_UNDEF_VAR		[<name>, <val>] where <name> is the name of the global variable not
;														defined and <val> is the value that should be assigned
;		if @error = $IU_VAR_NOT_FOUND	[<sect>, <name>] where <sect> is the .ini section where the variable
;														<name> has not been found
;---
	local $tmp
	local enum $_IU_SECT, $_IU_INI_VAR, $_IU_GLOB_VAR, $_IU_DEF		; Positions in inner arrays
	
	$tmp = UBound($def, 2)
	if ($tmp <> 4) then
		local $res[2] = ["InnerDim", $tmp]
		SetError($IU_INVALID_DEF)
		return $res
	endif
	for $n = 0 to (UBound($def) - 1)
		$tmp = IniRead($ini, $def[$n][$_IU_SECT], $def[$n][$_IU_INI_VAR], $def[$n][$_IU_DEF])
		if ($tmp == default) then
			local $res[2] = [$def[$n][$_IU_SECT], $def[$n][$_IU_INI_VAR]]
			SetError($IU_VAR_NOT_FOUND)
			return $res
		endif
		if (not Assign($def[$n][$_IU_GLOB_VAR], $tmp, 6)) then	; 2 + 4 = global scope + fail if not existant
			local $res[2] = [$def[$n][$_IU_GLOB_VAR], $tmp]
			SetError($IU_UNDEF_VAR)
			return $res
		endif
	next
	return 1
endfunc



func IUWriteArrays(const $ini, const byref $def)
;---
; Writes to $ini file, values from global arrays as specified in $def; on errors, @error is set and details
; are returned in an array.
;
; $ini	ini file path
; $def	array with definitions of .ini variables and global arrays; every entry is an inner array in the form:
;			[IniSectionName, IniVarNameStruc, GlobalArrayName]
;		where:
;			IniSectionName	name of the .ini section to be used
;			IniVarNameStruc	structure of names of the .ini variables inside IniSectionName (details below)
;			GlobalArrayName	name of the global array to be written (must already exists)
;		If $def has uncorrect dimensions or name structure, the function sets @error to $IU_INVALID_DEF and
;		returns.
;		If GlobalArrayName has not been defined, the function sets @error to $IU_UNDEF_VAR and returns.
;		The entry IniVarNameStruc is list of elements separater by "=" ("=" is an unusable char in .ini
;		variable name); it can have 2 forms:
;			"List=VarName1=VarName2=VarName3=..." where VarName<n> are variable names that should be written
;					in appearing order and assigned from array elements 0..<n-1>; the list should end with
;					a VarName, not with a "="
;			"Step=VarNamePrefix=DigitsNum=StartNum=EndNum" where VarNamePrefix is the first part of variable
;					name at which is appended a number composed by DigitsNum digits, starting at StartNum
;					value and ending at EndNum value included
;		Array elements written are only those specified; if an array is longer than variable names, the tail
;		is ignored, while an array shorter or a variable that is not an array is an error.
;		On errors on IniVarNameStruc structure, the function sets @error to $IU_INVALID_DEF and returns.
; ret	if no error occurred			1
;		if @error = $IU_INVALID_DEF		["InnerDim", <n>] where <n> is used dims of inner array (should be 4)
;										["TooShort", <str>] where <str> is the invalid structure used
;										["InvStruct", <str>] where <str> is the invalid structure used
;										["InvOper", <oper>] where <oper> is the invalid operation (not "List",
;															nor "Step")
;										["InvArray", <name>] where <name> is the name of variable that is not
;															an array or is an array but not with one dimension
;															or without enough items
;		if @error = $IU_UNDEF_VAR		[<name>, ""] where <name> is the global variable name not defined
;		if @error = $IU_WRITE_FAILED	[<sect>, <name>] where <sect> is the .ini section where the variable
;														<name> has not been written
;---
	local $tmp, $one_def, $elem_count, $oper, $n, $i, $glob_arr_name, $tmp_arr, $ini_var_name
	local $prefix, $digits, $start, $end
	local enum $_IU_SECT, $_IU_INI_ARR, $_IU_GLOB_ARR				; Positions in inner arrays

	$tmp = UBound($def, 2)
	if ($tmp <> 3) then
		local $res[2] = ["InnerDim", $tmp]
		SetError($IU_INVALID_DEF)
		return $res
	endif
	for $n = 0 to (UBound($def) - 1)
		$one_def = StringSplit($def[$n][$_IU_INI_ARR], "=")
		$elem_count = $one_def[0]
		if ($elem_count == 1) then
			local $res[2] = ["TooShort", $def[$n][1]]
			SetError($IU_INVALID_DEF)
			return $res
		endif
		$glob_arr_name = $def[$n][$_IU_GLOB_ARR]
		if (IsDeclared($glob_arr_name) <> 1) then
			local $res[2] = [$glob_arr_name, ""]
			SetError($IU_UNDEF_VAR)
			return $res
		endif
		$tmp_arr = Eval($glob_arr_name)
		if (UBound($tmp_arr, 0) <> 1) then
			local $res[2] = ["InvArray", $glob_arr_name]
			SetError($IU_INVALID_DEF)
			return $res
		endif
		$oper = StringLower($one_def[1])
		switch $oper
			case "list"
				if ($elem_count < 2) then
					local $res[2] = ["TooShort", $def[$n][1]]
					SetError($IU_INVALID_DEF)
					return $res
				endif
				if (UBound($tmp_arr) < ($elem_count - 1)) then
					local $res[2] = ["InvArray", $glob_arr_name]
					SetError($IU_INVALID_DEF)
					return $res
				endif
				for $i = 2 to $elem_count
					$ini_var_name = $one_def[$i]
					if (not IniWrite($ini, $def[$n][$_IU_SECT], $ini_var_name, $tmp_arr[$i - 2])) then
						local $res[2] = [$def[$n][$_IU_SECT], $ini_var_name]
						SetError($IU_WRITE_FAILED)
						return $res
					endif
				next
			case "step"
				if ($elem_count <> 5) then
					local $res[2] = ["InvStruct", $def[$n][1]]
					SetError($IU_INVALID_DEF)
					return $res
				endif
				$prefix = $one_def[2]
				$digits = $one_def[3]
				$start = $one_def[4]
				$end = $one_def[5]
				if (UBound($tmp_arr) < ($end - $start + 1)) then
					local $res[2] = ["InvArray", $glob_arr_name]
					SetError($IU_INVALID_DEF)
					return $res
				endif
				for $i = $start to $end
					$ini_var_name = StringFormat("%s%0" & $digits & "u", $prefix, $i)
					if (not IniWrite($ini, $def[$n][$_IU_SECT], $ini_var_name, $tmp_arr[$i - $start])) then
						local $res[2] = [$def[$n][$_IU_SECT], $ini_var_name]
						SetError($IU_WRITE_FAILED)
						return $res
					endif
				next
			case else
				local $res[2] = ["InvOper", $oper]
				SetError($IU_INVALID_DEF)
				return $res
		endswitch
	next
	return 1
endfunc



func IUWriteVars(const $ini, const byref $def)
;---
; Writes to $ini file, values from global variables as specified in $def; on errors, @error is set and
; details are returned in an array.
;
; $ini	ini file path
; $def	array with definitions of .ini and global variables; every entry is an inner array in the form:
;			[IniSectionName, IniVarName, GlobalVarName]
;		where:
;			IniSectionName	name of the .ini section to be used
;			IniVarName		name of the .ini variable inside IniSectionName
;			GlobalVarName	name of the global variable to be assigned (must already exists)
;		If $def has uncorrect dimensions, the function sets @error to $IU_INVALID_DEF and returns.
;		If GlobalVarName has not been defined, the function sets @error to $IU_UNDEF_VAR and returns.
; ret	if no error occurred			1
;		if @error = $IU_INVALID_DEF		["InnerDim", <n>] where <n> is used dims of inner array (should be 3)
;		if @error = $IU_UNDEF_VAR		[<name>, ""] where <name> is the global variable name not defined
;		if @error = $IU_WRITE_FAILED	[<sect>, <name>] where <sect> is the .ini section where the variable
;														<name> has not been written
;---
	local $tmp
	local enum $_IU_SECT, $_IU_INI_VAR, $_IU_GLOB_VAR				; Positions in inner arrays
	
	$tmp = UBound($def, 2)
	if ($tmp <> 3) then
		local $res[2] = ["InnerDim", $tmp]
		SetError($IU_INVALID_DEF)
		return $res
	endif
	for $n = 0 to (UBound($def) - 1)
		if (IsDeclared($def[$n][$_IU_GLOB_VAR]) <> 1) then
			local $res[2] = [$def[$n][$_IU_GLOB_VAR], ""]
			SetError($IU_UNDEF_VAR)
			return $res
		endif
		if (not IniWrite($ini, $def[$n][$_IU_SECT], $def[$n][$_IU_INI_VAR], Eval($def[$n][$_IU_GLOB_VAR]))) then
			local $res[2] = [$def[$n][$_IU_SECT], $def[$n][$_IU_INI_VAR]]
			SetError($IU_WRITE_FAILED)
			return $res
		endif
	next
	return 1
endfunc



;--- Private functions code ------------------------------------------------------------------------



func _IUAddTrailingBS($d)
;---
; Adds a trailing backslash if $d has not already (i.e. on root dir.).
;
; $d	directory where add tralinig backslash
; ret	directory $d with a trailing backslash
;---
	if (StringRight($d, 1) == "\") then
		return $d
	else
		return $d & "\"
	endif
endfunc



;--- Initialization code constants -----------------------------------------------------------------



;--- Initialization code variables -----------------------------------------------------------------



;--- Initialization code ---------------------------------------------------------------------------


